<?php
/**
 * Rest API functions.
 *
 * @since 2.5.0
 *
 * @package GP Premium
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // No direct access, please.
}

/**
 * Font library REST API endpoints class.
 */
class GeneratePress_Font_Library_Rest extends WP_REST_Controller {
	/**
	 * Instance.
	 *
	 * @access private
	 * @var object Instance
	 */
	private static $instance;

	/**
	 * Namespace.
	 *
	 * @var string
	 */
	protected $namespace = 'generatepress-font-library/v';

	/**
	 * Version.
	 *
	 * @var string
	 */
	protected $version = '1';

	/**
	 * Initiator.
	 *
	 * @return object initialized object of class.
	 */
	public static function get_instance() {
		if ( ! isset( self::$instance ) ) {
			self::$instance = new self();
		}

		return self::$instance;
	}

	/**
	 * GenerateBlocks_Rest constructor.
	 */
	public function __construct() {
		add_action( 'rest_api_init', array( $this, 'register_routes' ) );
	}

	/**
	 * Register rest routes.
	 */
	public function register_routes() {
		$namespace = $this->namespace . $this->version;

		// Get fonts from CPT.
		register_rest_route(
			$namespace,
			'/get-fonts/',
			array(
				'methods'  => WP_REST_Server::READABLE,
				'callback' => array( $this, 'get_fonts' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		// Download a Google font.
		register_rest_route(
			$namespace,
			'/download-google-font/',
			array(
				'methods'             => WP_REST_Server::EDITABLE,
				'callback'            => array( $this, 'download_google_font' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		// Upload a custom font.
		register_rest_route(
			$namespace,
			'/upload-custom-font/',
			array(
				'methods'             => WP_REST_Server::EDITABLE,
				'callback'            => array( $this, 'upload_custom_font' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		// Delete a font family.
		register_rest_route(
			$namespace,
			'/delete-font/',
			array(
				'methods'             => WP_REST_Server::EDITABLE,
				'callback'            => array( $this, 'delete_font' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		// Get font library settings.
		register_rest_route(
			$namespace,
			'/get-settings/',
			array(
				'methods'  => WP_REST_Server::READABLE,
				'callback' => array( $this, 'get_settings' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		// Get font library settings.
		register_rest_route(
			$namespace,
			'/set-settings/',
			array(
				'methods'             => WP_REST_Server::EDITABLE,
				'callback'            => array( $this, 'set_settings' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);

		register_rest_route(
			$namespace,
			'/update-font-post/',
			array(
				'methods'             => WP_REST_Server::EDITABLE,
				'callback'            => array( $this, 'update_font_post' ),
				'permission_callback' => array( $this, 'edit_posts_permission' ),
			)
		);
	}

	/**
	 * Get font posts.
	 *
	 * @param WP_REST_Request $request The request object.
	 *
	 * @return array The response.
	 */
	public function get_fonts( WP_REST_Request $request ) {
		$name     = $request->get_param( 'name' );
		$response = GeneratePress_Font_Library::get_fonts( $name );

		return $this->success( $response );
	}

	/**
	 * Generate font CSS.
	 *
	 * @return mixed
	 */
	public function build_css_file() {

		$result = GeneratePress_Font_Library::build_css_file();

		if ( is_wp_error( $result ) ) {
			return $this->error( 'font_css_generation_failed', __( 'Failed to generate font CSS.', 'gp-premium' ) );
		}

		return $this->success( $result );
	}

	/**
	 * Delete a specific font from the library and the associated CPT post.
	 *
	 * @param WP_REST_Request $request request object.
	 *
	 * @return mixed
	 */
	public function delete_font( WP_REST_Request $request ) {
		$font_id        = $request->get_param( 'fontId' );
		$slug           = get_post_field( 'post_name', $font_id );
		$upload_dir     = wp_get_upload_dir();
		$font_base_path = trailingslashit( $upload_dir['basedir'] ) . 'generatepress/fonts/' . $slug . '/';

		// Delete the font post.
		$success = wp_delete_post( $font_id, true );

		if ( ! $success ) {
			return $this->error(
				'font_post_delete_failed',
				__( 'Failed to delete font post.', 'gp-premium' )
			);
		}

		// Delete the font sub folder if it exists.
		if ( file_exists( $font_base_path ) ) {
			GeneratePress_Font_Library::delete_directory( $font_base_path );
		}

		// Regenerate the font CSS.
		$this->build_css_file();

		// Return success.
		return $this->success( __( 'Font successfully deleted!', 'gp-premium' ) );
	}

	/**
	 * Download a specific Google font and update the CPT.
	 *
	 * @param WP_REST_Request $request request object.
	 *
	 * @return mixed
	 */
	public function download_google_font( WP_REST_Request $request ) {
		$font          = $request->get_param( 'font' );
		$variants      = $request->get_param( 'variants' );
		$font_settings = $font['font_family_settings'];
		$font_family   = explode( ', ', $font_settings['fontFamily'] );

		// Remove the main font-family.
		array_shift( $font_family );

		$fallback = implode( ', ', $font_family );

		// Check for existing font post, and create new one if not found.
		$args = array(
			'name'                   => $font_settings['slug'], // Font slug.
			'post_type'              => GeneratePress_Font_Library::FONT_LIBRARY_CPT,
			'post_status'            => 'publish',
			'numberposts'            => 1,
			'fields'                 => 'ids',
			'no_found_rows'          => true,
			'update_post_meta_cache' => false,
			'update_term_meta_cache' => false,
		);
		$font_posts = get_posts( $args );

		if ( ! empty( $font_posts ) ) {
			$font_post = $font_posts[0];
			$formatted_variants = get_post_meta( $font_post, 'gp_font_variants', true );
			if ( ! is_array( $formatted_variants ) ) {
				$formatted_variants = array();
			}
		} else {
			$formatted_variants = array();
		}

		$upload_dir = wp_get_upload_dir();
		$base_path  = trailingslashit( $upload_dir['basedir'] ) . 'generatepress/fonts/' . $font_settings['slug'] . '/';

		// Ensure the directory exists.
		if ( ! file_exists( $base_path ) ) {
			wp_mkdir_p( $base_path );
		}

		// Download each font variant.
		foreach ( $variants as $variant ) {
			$parsed_url  = wp_parse_url( $variant['src'], PHP_URL_PATH );
			$file_name   = basename( $parsed_url );
			$file_path   = $base_path . $file_name;
			$file_url    = trailingslashit( $upload_dir['baseurl'] ) . 'generatepress/fonts/' . $font_settings['slug'] . '/' . $file_name;

			$formatted_variants = GeneratePress_Font_Library::check_variants(
				$formatted_variants,
				array(
					'src'        => $file_url,
					'name'       => GeneratePress_Font_Library::get_variant_name( $variant ),
					'fontFamily' => $variant['fontFamily'],
					'fontStyle'  => $variant['fontStyle'],
					'fontWeight' => $variant['fontWeight'],
					'preview'    => $variant['preview'],
					'isVariable' => false,
					'disabled'   => false,
					'source'     => 'google',
				),
				$base_path
			);

			// Download the file.
			$response = wp_remote_get( $variant['src'] );

			if ( is_wp_error( $response ) ) {
				$error_message = $response->get_error_message();
				return $this->failed( "Failed to download {$variant['fontFamily']} from {$variant['src']}: $error_message" );
			}

			// Save the file.
			$filesystem = generate_premium_get_wp_filesystem();

			if ( ! $filesystem ) {
				return $this->failed( 'Error setting up the file system object.' );
			}

			$file_contents = wp_remote_retrieve_body( $response );

			if ( ! $file_contents ) {
				return $this->failed( "Failed to download $variant from {$variant['src']}: Empty body" );
			} else {
				// Assuming $filesystem is already set up correctly.
				$chmod_file = defined( 'FS_CHMOD_FILE' ) ? FS_CHMOD_FILE : 0644;

				if ( is_writable( $file_path ) || is_writable( dirname( $file_path ) ) ) {
					if ( ! $filesystem->put_contents( $file_path, $file_contents, $chmod_file ) ) {
						return $this->failed( "Failed to download $variant from {$variant['src']}." );
					}
				}
			}
		}

		if ( ! empty( $font_posts ) ) {
			$font_post = $font_posts[0];
		} else {
			// No post found with that slug so create new post.
			$font_post = wp_insert_post(
				array(
					'post_title'  => $font_settings['name'],
					'post_name'   => $font_settings['slug'], // Font slug.
					'post_type'   => GeneratePress_Font_Library::FONT_LIBRARY_CPT,
					'post_status' => 'publish',
					'wp_error'    => true,
				)
			);
		}

		if ( ! is_wp_error( $font_post ) ) {
			$font_slug = get_post_field( 'post_name', $font_post );
			// Update the font post meta with new variants, overwriting existing font variant meta.
			update_post_meta( $font_post, 'gp_font_variants', $formatted_variants );

			$font_display = get_post_meta( $font_post, 'gp_font_display', true );
			if ( empty( $font_display ) ) {
				$font_display = 'auto';
			}

			update_post_meta( $font_post, 'gp_font_display', $font_display );
			update_post_meta( $font_post, 'gp_font_source', 'google' );
			update_post_meta( $font_post, 'gp_font_fallback', $fallback );
			update_post_meta( $font_post, 'gp_font_preview', $font_settings['preview'] );

			// Set the initial CSS variable.
			$css_variable = GeneratePress_Font_Library::CSS_VAR_PREFIX . $font_slug;
			update_post_meta( $font_post, 'gp_font_variable', $css_variable );

			// Generate the font CSS.
			$this->build_css_file();
		} else {
			return $this->error(
				'font_post_create_failed',
				__( 'Failed to create font post.', 'gp-premium' )
			);
		}

		// Return success.
		return $this->success( __( 'All font variations successfully downloaded!', 'gp-premium' ) );
	}

	/**
	 * Upload a specific custom font and update the CPT.
	 *
	 * @param WP_REST_Request $request request object.
	 *
	 * @return mixed
	 */
	public function upload_custom_font( WP_REST_Request $request ) {
		$settings = json_decode( $request->get_param( 'font_face_settings' ) );

		// phpcs:disable WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase
		$font_family  = $settings->fontFamily;
		$font_style   = $settings->fontStyle;
		$font_weight  = $settings->fontWeight;
		$is_variable  = $settings->isVariable;
		// phpcs:enable WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase

		$font_slug    = $request->get_param( 'font_slug' );
		$file_params  = $request->get_file_params();

		// Move the uploaded font asset from the temp folder to the fonts directory.
		if ( ! function_exists( 'wp_handle_upload' ) ) {
			require_once ABSPATH . 'wp-admin/includes/file.php';
		}

		$file      = $file_params[ $settings->src ];
		$font_file = GeneratePress_Font_Library::handle_font_file_upload( $file, $font_slug );
		if ( is_wp_error( $font_file ) ) {
			return $this->error( 'font_file_upload_failed', $font_file->get_error_message() );
		}

		// Check for existing font post, and create new one if not found.
		$args = array(
			'name'                   => $font_slug, // Font slug.
			'post_type'              => GeneratePress_Font_Library::FONT_LIBRARY_CPT,
			'fields'                 => 'ids', // return post ID's not full post objects.
			'post_status'            => 'publish',
			'numberposts'            => 1,
			'no_found_rows'          => true,
			'update_post_meta_cache' => false,
			'update_term_meta_cache' => false,
		);
		$font_posts = get_posts( $args );

		if ( ! empty( $font_posts ) ) {
			$font_post = $font_posts[0];
		} else {
			// No post found with that slug so create new post.
			$font_post = wp_insert_post(
				array(
					'post_title'  => $font_family,
					'post_name'   => $font_slug,
					'post_type'   => GeneratePress_Font_Library::FONT_LIBRARY_CPT,
					'post_status' => 'publish',
					'wp_error'    => true,
				)
			);
		}

		if ( ! is_wp_error( $font_post ) ) {
			$variants = get_post_meta( $font_post, 'gp_font_variants', true );
			if ( ! is_array( $variants ) ) {
				$variants = array();
			}

			$font_display = get_post_meta( $font_post, 'gp_font_display', true );

			// Prepare new variant from settings.
			$name = GeneratePress_Font_Library::get_variant_name(
				(array) $settings
			);

			$upload_dir = wp_get_upload_dir();
			$base_path  = trailingslashit( $upload_dir['basedir'] ) . 'generatepress/fonts/' . $font_slug . '/'; // phpcs:ignore WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase

			$checked_variants = GeneratePress_Font_Library::check_variants(
				$variants,
				array(
					'src'        => $font_file['url'],
					'fontFamily' => $font_family,
					'fontStyle'  => $font_style,
					'fontWeight' => $font_weight,
					'name'       => $name,
					'isVariable' => $is_variable,
					'source'     => 'custom',
					'disabled'   => false,
					'preview'    => '',
				),
				$base_path
			);

			// Update the font post meta with merged variants.
			update_post_meta( $font_post, 'gp_font_variants', $checked_variants );

			if ( empty( $font_display ) ) {
				$font_display = 'auto';
			}
			update_post_meta( $font_post, 'gp_font_display', $font_display );
			update_post_meta( $font_post, 'gp_font_source', 'custom' );

			// Set the initial CSS variable.
			$css_variable = GeneratePress_Font_Library::CSS_VAR_PREFIX . $font_slug;
			update_post_meta( $font_post, 'gp_font_variable', $css_variable );

			// Generate the font CSS.
			$generate_css = $this->build_css_file();

			if ( false === $generate_css->success ) {
				return $this->error( 500, __( 'CSS Generation failed', 'gp-premium' ) );
			}
		} else {
			return $this->error( 500, __( 'Failed to create font post.', 'gp-premium' ) );
		}

		// Return success.
		return $this->success( __( 'All font variations successfully downloaded!', 'gp-premium' ) );
	}

	/**
	 * Get font library settings. At the moment this is just the Google GDPR setting.
	 *
	 * @return mixed
	 */
	public function get_settings() {
		$user_id = get_current_user_id();
		$google_gdpr = (bool) get_user_meta( $user_id, 'gp_font_library_google_gdpr', true );
		return $this->success( array( 'google_gdpr' => $google_gdpr ) );
	}

	/**
	 * Update font library settings. At the moment this is just the Google GDPR setting.
	 *
	 * @param WP_REST_Request $request request object.
	 *
	 * @return mixed
	 */
	public function set_settings( WP_REST_Request $request ) {
		$google_gdpr_value = (bool) $request->get_param( 'settings' )['google_gdpr'];
		$user_id = get_current_user_id();
		update_user_meta( $user_id, 'gp_font_library_google_gdpr', $google_gdpr_value );
		$updated = (bool) get_user_meta( $user_id, 'gp_font_library_google_gdpr', true );

		// Return success.
		return $this->success(
			array(
				'message' => __( 'Font settings successfully updated!', 'gp-premium' ),
				'response' => $updated,
				'gdpr_input' => $google_gdpr_value,
			)
		);
	}

	/**
	 * Update a font post.
	 *
	 * @param WP_REST_Request $request request object.
	 *
	 * @return mixed
	 */
	public function update_font_post( WP_REST_Request $request ) {
		$font_id           = $request->get_param( 'id' );
		$status            = $request->get_param( 'status' );
		$font_family_alias = $request->get_param( 'alias' );
		$new_variants      = $request->get_param( 'newVariants' );
		$delete_variants   = $request->get_param( 'deleteVariants' );
		$font_display      = $request->get_param( 'fontDisplay' );
		$fallback          = $request->get_param( 'fallback' );
		$css_variable      = $request->get_param( 'cssVariable' );
		$slug              = get_post_field( 'post_name', $font_id );

		// Update the font post.
		wp_update_post(
			array(
				'ID'          => $font_id,
				'post_status' => $status,
				'meta_input'  => array(
					'gp_font_family_alias' => $font_family_alias,
					'gp_font_variants'     => $new_variants,
					'gp_font_display'      => $font_display,
					'gp_font_fallback'     => $fallback,
					'gp_font_variable'     => $css_variable,
				),
			)
		);

		$upload_dir = wp_get_upload_dir();
		$base_path  = trailingslashit( $upload_dir['basedir'] ) . 'generatepress/fonts/' . $slug . '/';
		foreach ( $delete_variants as $variant ) {
			if ( isset( $variant['deleteStatus'] ) && $variant['deleteStatus'] ) {
				$file_path = $base_path . basename( $variant['src'] );
				if ( file_exists( $file_path ) ) {
					unlink( $file_path );
				}
			}
		}

		// Regenerate the font CSS.
		$this->build_css_file();

		// Return success.
		return $this->success( __( 'Font post successfully updated!', 'gp-premium' ) );
	}

	/**
	 * Get edit options permissions.
	 *
	 * @return bool
	 */
	public function update_settings_permission() {
		return current_user_can( 'manage_options' );
	}

	/**
	 * Get edit posts permissions.
	 *
	 * @return bool
	 */
	public function edit_posts_permission() {
		return current_user_can( 'edit_posts' );
	}

	/**
	 * Success rest.
	 *
	 * @param mixed $response response data.
	 * @param mixed $data     data.
	 * @return mixed
	 */
	public function success( $response, $data = null ) {
		return new WP_REST_Response(
			array(
				'success'  => true,
				'response' => $response,
				'data'     => $data,
			),
			200
		);
	}

	/**
	 * Failed rest.
	 *
	 * @param mixed $response response data.
	 * @return mixed
	 */
	public function failed( $response ) {
		return new WP_REST_Response(
			array(
				'success'  => false,
				'response' => $response,
			),
			200
		);
	}

	/**
	 * Error rest.
	 *
	 * @param mixed $code     error code.
	 * @param mixed $response response data.
	 * @return mixed
	 */
	public function error( $code, $response ) {
		return new WP_REST_Response(
			array(
				'error'      => true,
				'success'    => false,
				'error_code' => $code,
				'response'   => $response,
			),
			500
		);
	}
}

GeneratePress_Font_Library_Rest::get_instance();
